﻿set-location c:;
$ErrorActionPreference = "Stop";

$restoreInfoTypeFound_38045c3a1f64442fbc69d710863e6e6c = $false;

#####################################################################################################
## New-CustomObjectsFromScriptInternal function
#####################################################################################################
function New-CustomObjectsFromScriptInternal
{
<#
.SYNOPSIS

Compiles a C# script.

.DESCRIPTION

Compiles a C# script typically containing the definitions of .NET objects to be used in functions in this module.
This function is intended for internal use only and should not generally be called by the end user.

.PARAMETER Script

The script to be compiled.

.INPUTS

None. You cannot pipe data to New-CustomObjectsFromScriptInternal.

.OUTPUTS

None. The resulting assembly will be loaded into memory.

.EXAMPLE

New-CustomObjectsFromScriptInternal -Script "public class Foo { public Foo() {} }";

#>
    param
    (
        [Parameter(Mandatory = $true)]
        [string] $Script
    )

    $parameters = New-Object System.CodeDom.Compiler.CompilerParameters;
    $parameters.CompilerOptions = "/target:library";
    $parameters.GenerateExecutable = $false;
    $parameters.GenerateInMemory = $true;
    $parameters.IncludeDebugInformation = $false;

    $provider = New-Object Microsoft.CSharp.CSharpCodeProvider;
    $results = $provider.CompileAssemblyFromSource($parameters, $Script);
    if ($null -eq $results)
    {
        throw "Compile operation failed.";
    }

    if ($null -ne $results.Errors -and $results.Errors.HasErrors -eq $true)
    {
        $errorMessage = "Error compiling script.";
        foreach($compilerError in $results.Errors)
        {
            $errorMessage += "`r`n" + $compilerError;
        }
        throw $errorMessage;
    }
}

#####################################################################################################
## Add-CustomTypesInternal function
#####################################################################################################
function Add-CustomTypesInternal
{
<#
.SYNOPSIS

Creates .NET types used elsewhere in this module.

.DESCRIPTION

Creates and loads the Sql.RestoreInfo and Sql.RestoreInfoList types used elsewhere in this module, if they do not already exist.

This function is intended for internal use only and should not generally be called by the end user.

.INPUTS

None. You cannot pipe data to Add-CustomTypesInternal.

.OUTPUTS

None. The resulting types will be created and loaded into memory.

.EXAMPLE

Add-CustomTypesInternal;

#>
    if ($restoreInfoTypeFound_38045c3a1f64442fbc69d710863e6e6c -eq $false)
    {
        $assemblies = [System.AppDomain]::CurrentDomain.GetAssemblies();
        foreach ($assembly in $assemblies | Where-Object { [System.String]::IsNullOrEmpty($_.Location) })
        {
            $types = $assembly.GetTypes();
            foreach($type in $types)
            {
                if ($type.FullName -eq "Sql.RestoreInfo")
                {
                    $restoreInfoTypeFound_38045c3a1f64442fbc69d710863e6e6c = $true;
                }
            }
        }
    }

    if ($restoreInfoTypeFound_38045c3a1f64442fbc69d710863e6e6c -eq $false)
    {
        ## This class needs to be C# 2.0 compliant
        $restoreInfoClass = @"
            using System;
            using System.Collections.Generic;
            namespace Sql
            {
                public class RestoreInfo
                {
                    private string _backupFileName;
                    private decimal _firstLsn;
                    private decimal _lastLsn;
                    private decimal _checkpointLsn;
                    private decimal _databaseBackupLsn;
                    private long _backupSize;
                    private DateTime _backupStartDate;
                    private DateTime _backupFinishDate;
                    private string _backupType;
                    private string _serverName;
                    private string _databaseName;
                    private bool _compressed;
                    private int _position;
                    private int _compatibilityLevel;
                    private string _encryption;
                    private bool _isCopyOnly;

                    public RestoreInfo()
                    {
                    }

                    public string BackupFileName { get { return _backupFileName; } set { _backupFileName = value; } }
                    public decimal FirstLSN { get { return _firstLsn; } set { _firstLsn = value; } }
                    public decimal LastLSN { get { return _lastLsn; } set { _lastLsn = value; } }
                    public decimal CheckpointLSN { get { return _checkpointLsn; } set { _checkpointLsn = value; } }
                    public decimal DatabaseBackupLSN { get { return _databaseBackupLsn; } set { _databaseBackupLsn = value; } }
                    public long BackupSize { get { return _backupSize; } set { _backupSize = value; } }
                    public DateTime BackupStartDate { get { return _backupStartDate; } set { _backupStartDate = value; } }
                    public DateTime BackupFinishDate { get { return _backupFinishDate; } set { _backupFinishDate = value; } }
                    public string BackupType { get { return _backupType; } set { _backupType = value; } }
                    public string ServerName { get { return _serverName; } set { _serverName = value; } }
                    public string DatabaseName { get { return _databaseName; } set { _databaseName = value; } }
                    public bool Compressed { get { return _compressed; } set { _compressed = value; } }
                    public int Position { get { return _position; } set { _position = value; } }
                    public int CompatibilityLevel { get { return _compatibilityLevel; } set { _compatibilityLevel = value; } }
                    public string Encryption { get { return _encryption; } set { _encryption = value; } }
                    public bool IsCopyOnly { get { return _isCopyOnly; } set { _isCopyOnly = value; } }
                }

                public class RestoreInfoList : List<RestoreInfo>
                {
                    public RestoreInfoList() : base() { }

                    public RestoreInfoList(IEnumerable<RestoreInfo> collection) : base(collection) { }

                    public RestoreInfoList(int size) : base(size) { }

                    private Dictionary<string, RestoreInfo> _index = null;

                    private void UpdateIndex()
                    {
                        _index = new Dictionary<string, RestoreInfo>();
                        foreach (RestoreInfo restoreInfo in this) { _index.Add(restoreInfo.BackupFileName, restoreInfo); }
                    }

                    public RestoreInfo GetRestoreInfo(string fileName)
                    {
                        if (_index == null) { UpdateIndex(); }
                        if (_index.ContainsKey(fileName)) { return _index[fileName]; }
                        else { return null; }
                    }
                }

                public class FileListOnlyInfo
                {
                    private string _backupFileName;
                    private string _logicalName;
                    private string _physicalName;
                    private string _backupFileType;
                    private string _fileGroupName;
                    private long _size;
                    private long _maxSize;
                    private long _fileID;
                    private decimal _createLsn;
                    private decimal? _dropLsn;
                    private Guid _uniqueId;
                    private decimal? _readOnlyLsn;
                    private decimal? _readWriteLsn;
                    private long _backupSizeInBytes;
                    private int _sourceBlockSize;
                    private int _fileGroupId;
                    private decimal? _differentialBaseLsn;
                    private Guid _differentialBaseGuid;
                    private bool _isReadOnly;
                    private bool _isPresent;
                    private byte[] _tdeThumbprint;

                    public FileListOnlyInfo()
                    {
                    }

                    public string BackupFileName { get { return _backupFileName; } set { _backupFileName = value; } }
                    public string LogicalName { get { return _logicalName; } set { _logicalName = value; } }
                    public string PhysicalName { get { return _physicalName; } set { _physicalName = value; } }
                    public string BackupFileType { get { return _backupFileType; } set { _backupFileType = value; } }
                    public string FileGroupName { get { return _fileGroupName; } set { _fileGroupName = value; } }
                    public long Size { get { return _size; } set { _size = value; } }
                    public long MaxSize { get { return _maxSize; } set { _maxSize = value; } }
                    public long FileID { get { return _fileID; } set { _fileID = value; } }
                    public decimal CreateLsn { get { return _createLsn; } set { _createLsn = value; } }
                    public decimal? DropLsn { get { return _dropLsn; } set { _dropLsn = value; } }
                    public Guid UniqueId { get { return _uniqueId; } set { _uniqueId = value; } }
                    public decimal? ReadOnlyLsn { get { return _readOnlyLsn; } set { _readOnlyLsn = value; } }
                    public decimal? ReadWriteLsn { get { return _readWriteLsn; } set { _readWriteLsn = value; } }
                    public long BackupSizeInBytes { get { return _backupSizeInBytes; } set { _backupSizeInBytes = value; } }
                    public int SourceBlockSize { get { return _sourceBlockSize; } set { _sourceBlockSize = value; } }
                    public int FileGroupId { get { return _fileGroupId; } set { _fileGroupId = value; } }
                    public decimal? DifferentialBaseLsn { get { return _differentialBaseLsn; } set { _differentialBaseLsn = value; } }
                    public Guid DifferentialBaseGuid { get { return _differentialBaseGuid; } set { _differentialBaseGuid = value; } }
                    public bool IsReadOnly { get { return _isReadOnly; } set { _isReadOnly = value; } }
                    public bool IsPresent { get { return _isPresent; } set { _isPresent = value; } }
                    public byte[] TdeThumbprint { get { return _tdeThumbprint; } set { _tdeThumbprint = value; } }
                }

                public class FileListOnlyInfoList : List<FileListOnlyInfo>
                {
                    public FileListOnlyInfoList() : base() { }

                    public FileListOnlyInfoList(IEnumerable<FileListOnlyInfo> collection) : base(collection) { }

                    public FileListOnlyInfoList(int size) : base(size) { }

                    private Dictionary<string, FileListOnlyInfo> _indexByLogicalName = null;

                    private void UpdateIndex()
                    {
                        _indexByLogicalName = new Dictionary<string, FileListOnlyInfo>();
                        foreach(FileListOnlyInfo fileListOnlyInfo in this) { _indexByLogicalName.Add(fileListOnlyInfo.LogicalName, fileListOnlyInfo); }
                    }

                    public FileListOnlyInfo GetFileListOnlyInfo(string logicalName)
                    {
                        if (_indexByLogicalName == null) { UpdateIndex(); }
                        if (_indexByLogicalName.ContainsKey(logicalName)) { return _indexByLogicalName[logicalName]; }
                        else { return null; }
                    }
                }
            }
"@;

        New-CustomObjectsFromScriptInternal $restoreInfoClass;
        $restoreInfoTypeFound_38045c3a1f64442fbc69d710863e6e6c = $true;
    }
}

####################################################################################################
## Get-FileSize
####################################################################################################
function Get-FileSize
{
<#
.SYNOPSIS

Returns the size in bytes of a file.

.DESCRIPTION

Gets and return the size in bytes of a file if the file exists.

.PARAMETER FileName

The name of the file on which to obtain the file size.

.INPUTS

None. You cannot pipe data to Get-FileSize.

.OUTPUTS

The size in bytes of the file. The function returns zero if the file does not exist.

.EXAMPLE

Get-FileSize -FileName "c:\temp\data.txt";

#>
    param
    (
        [Parameter(Mandatory = $true)]
        [string]$FileName
    )

    if ([System.IO.File]::Exists($FileName) -eq $false)
    {
        return 0L;
    }

    $file = New-Object System.IO.FileInfo($FileName);
    return $file.Length;
}

#####################################################################################################
## Write-LogMessage
#####################################################################################################
function Write-LogMessage
{
<#
.SYNOPSIS

Writes a message to text log file.

.DESCRIPTION

Appends a string message to the end of a text log file.

.PARAMETER Message

A string that will be logged to the file.

.PARAMETER LogFileName

The name of the text log file.

.INPUTS

None. You cannot pipe data to Write-LogMessage.

.OUTPUTS

None.

.EXAMPLE

Write-LogMessage -Message "The operation succeeded." -LogFile "c:\temp\log.txt";

#>
    param
    (
        [Parameter(Mandatory = $true)]
        [string] $Message,

        [Parameter(Mandatory = $true)]
        [string] $LogFileName
    )

    $fs = New-Object System.IO.FileStream(
        $LogFileName, 
        [System.IO.FileMode]::Append, 
        [System.IO.FileAccess]::Write, 
        [System.IO.FileShare]::Read);

    $writer = New-Object System.IO.StreamWriter($fs);
    $writer.WriteLine("{0:yyyy-MM-dd HH:mm:ss.fff}: {1}", (Get-Date), $Message);
    $writer.Close();
}

#####################################################################################################
## Get-DataTableFromQuery
#####################################################################################################
function Get-DataTableFromQuery
{
<#
.SYNOPSIS

Loads a System.Data.DataTable from SQL Server.

.DESCRIPTION

Creates and populates a System.Data.DataTable from SQL Server based on a supplied query.

This function is primarily intended for internal use. However, the user may call this convenience method if desired.

.PARAMETER SqlConnection

A connected and open System.Data.SqlClient.SqlConnection object.

.PARAMETER Sql

The text of the SQL query to be executed.

.PARAMETER Parameters

An optional array of System.Data.SqlClient.SqlParameter objects that supply parameterized values to the sQL query.

.PARAMETER CommandTimeoutSeconds

An optional timeout value in seconds that will be applied. An exception will be generated if the query does not
complete in this amount of time. The default value is 3600 seconds (one hour).

.INPUTS

None. You cannot pipe data to Get-DataTableFromQuery.

.OUTPUTS

A System.Data.DataTable object containing the results from executing the query.

.EXAMPLE

Get-DataTableFromQuery -SqlConnection $connection -Sql "select * from MyTable;";

.EXAMPLE

Get-DataTableFromQuery -SqlConnection $connection -Sql "select * from MyTable where ID = @ID;" -Parameters @($idParameter) -CommandTimeoutSeconds 300;

#>
    param
    (
        [Parameter(Mandatory = $true)]
        [System.Data.SqlClient.SqlConnection] $SqlConnection,

        [Parameter(Mandatory = $true)]
        [string] $Sql,

        [System.Data.SqlClient.SqlParameter[]] $Parameters,
        [int] $CommandTimeoutSeconds = 3600
    )

    if ($SqlConnection -eq $null)
    {
        throw "SqlConnection parameter cannot be null.";
    }

    if ([System.String]::IsNullOrEmpty($Sql))
    {
        throw "Sql parameter cannot be empty or null.";
    }

    $table = New-Object System.Data.DataTable;
    try
    {
        try
        {
            $sqlCommand = New-Object System.Data.SqlClient.SqlCommand($Sql, $SqlConnection);
            $sqlCommand.CommandTimeout = $CommandTimeoutSeconds;
            if ($null -ne $Parameters)
            {
                foreach($parameter in $Parameters)
                {
                    [void]$sqlCommand.Parameters.Add($parameter);
                }
            }

            try
            {
                $adapter = New-Object System.Data.SqlClient.SqlDataAdapter($sqlCommand);
                [void]$adapter.Fill($table);
            }
            finally
            {
                if ($null -ne $adapter)
                {
                    $adapter.Dispose();
                }
            }
        }
        finally
        {
            if ($null -ne $sqlCommand)
            {
                $sqlCommand.Dispose();
            }
        }
    }
    catch [Exception]
    {
        if ($null -ne $table)
        {
            $table.Dispose();
        }
        throw;
    }

    ## Due to a PowerShell idiosyncracy, return $table inside an array
    return ,$table;
}

####################################################################################################
## Copy-FileWithRestart
####################################################################################################
function Copy-FileWithRestart
{
<#
.SYNOPSIS

Copy a file with restart logic.

.DESCRIPTION

This function performs a buffered copy operation on a file.

If the destination file exists and is smaller than source file, this function assumes that a
previous attempt to copy the file failed. It will then resume the copy operation at
approximately the point where the previous operation completed. (See also an impportant 
warning in the NOTES section.)

This function is particularly useful when copying large files across a network that may
be unreliable. See the Copy-FileWithRetry function which wraps Copy-FileWithRestart by
attempting a file copy operation a given number of times.

.PARAMETER SourceFileName

The name of the file to be copied.

.PARAMETER DestinationFileName

The destination location where the file will be copied.

.INPUTS

None. You cannot pipe data to Copy-FileWithRestart.

.OUTPUTS

None.

.NOTES

DATA LOSS WARNING: If the destination file already exists, this function makes the presumption
that the file is a previous failed attempt to copy the files. THIS ASSUMPTION IS NOT VALIDATED.
Additionally, this function may OVERWRITE the end of the existing file with data from source
file. Data loss will likely occur if the destination file is not in fact a result of a previous 
attempt to copy the source file.

.EXAMPLE

Copy-FileWithRestart -SourceFileName "c:\temp\file.txt" -DestinationFileName "d:\temp\filecopy.txt";

#>
    param
    (
        [Parameter(Mandatory = $true)]
        [string] $SourceFileName,

        [Parameter(Mandatory = $true)]
        [string] $DestinationFileName
    )

    if ([System.IO.File]::Exists($SourceFileName) -eq $false)
    {
        throw "Source file $SourceFileName does not exist.";
    }

    $bufferSize = 1048576;
    [long]$copyOffset = 0;

    if ([System.IO.File]::Exists($DestinationFileName))
    {
        $destSize = Get-FileSize($DestinationFileName);
        if ((Get-FileSize($SourceFileName)) -eq $destSize)
        {
            return;
        }

        $copyOffset = $destSize - ($destSize % $bufferSize);
        if ($copyOffset -gt $bufferSize)
        {
            $copyOffset -= $bufferSize;
        }
    }

    ##Write-LogMessage -LogFileName $logFileName -Message "Opening $SourceFilename for input.";
    $input = New-Object System.IO.FileStream ($SourceFileName, [System.IO.FileMode]::Open, [System.IO.FileAccess]::Read, [System.IO.FileShare]::Read);
    if ($input -eq $null)
    {
        throw "Unable to read from the source file $SourceFilename";
        return;
    }

    ##Write-LogMessage -LogFileName $logFileName -Message "Opening $DestinationFileName for output.";
    $output = New-Object System.IO.FileStream ($DestinationFileName, [System.IO.FileMode]::OpenOrCreate, [System.IO.FileAccess]::Write, [System.IO.FileShare]::Read);
    if ($output -eq $null)
    {
        throw "Unable to write to the desination file $DestinationFileName";
        return;
    }

    ##Write-LogMessage -LogFileName $logFileName -Message "Starting copy at offset $copyOffset.";
    $input.Position = $copyOffset;
    $output.Position = $copyOffset;

    ##Write-LogMessage -LogFileName $logFileName -Message "Initializing buffer.";
    $buffer = New-Object System.Byte[] $bufferSize;
    $bytesRead = 0;

    ##Write-LogMessage -LogFileName $logFileName -Message "Beginning copy.";
    try
    {
        do
        {
            try
            {
                $bytesRead = $input.Read($buffer, 0, $bufferSize);
                if ($bytesRead -gt 0)
                {
                    $output.Write($buffer, 0, $bytesRead);
                }
            }
            catch [System.Exception]
            {
                $ex = $_.Exception;
                throw ("Unexpected error while copying file data: " + $ex.Message);
                return;
            }
        }
        while ($bytesRead -gt 0);
    }
    catch
    {
        throw;
    }
    finally
    {
        $output.Close();
        $input.Close();
    }

    [System.IO.File]::SetCreationTime($DestinationFileName, [System.IO.FIle]::GetCreationTime($SourceFileName));
    [System.IO.File]::SetLastWriteTime($DestinationFileName, [System.IO.File]::GetLastWriteTime($SourceFileName));    
}

####################################################################################################
## Copy-FileWithRetry
####################################################################################################
function Copy-FileWithRetry
{
<#
.SYNOPSIS

Copy a file with retry logic.

.DESCRIPTION

This function performs a buffered copy operation on a file. If the copy fails, it will be
retried up to a specified number of times.

If the destination file exists and is smaller than source file, this function assumes that a
previous attempt to copy the file failed. It will then resume the copy operation at
approximately the point where the previous operation completed. (See also an impportant 
warning in the NOTES section.)

This function is particularly useful when copying large files across a network that may
be unreliable.

.PARAMETER SourceFileName

The name of the file to be copied.

.PARAMETER DestinationFileName

The destination location where the file will be copied.

.PARAMETER TryCount

The number of times to attempt the copy operation.

.PARAMETER RetryDelayMilliseconds

If the copy operation fails, specifies the amount of time in milliseconds to wait before
attempting another copy.

.INPUTS

None. You cannot pipe data to Copy-FileWithRetry.

.OUTPUTS

None.

.NOTES

DATA LOSS WARNING: If the destination file already exists, this function makes the presumption
that the file is a previous failed attempt to copy the files. THIS ASSUMPTION IS NOT VALIDATED.
Additionally, this function may OVERWRITE the end of the existing file with data from source
file. Data loss will likely occur if the destination file is not in fact a result of a previous 
attempt to copy the source file.

.EXAMPLE

Copy-FileWithRetry -SourceFileName "c:\temp\file.txt" -DestinationFileName "d:\temp\filecopy.txt" -TryCount 3 -RetryDelayMilliseconds 10000;

This command copies c:\temp\file.txt to d:\temp\filecopy.txt. If the copy fails, it will be 
attempted two more times (for a total of three tries) with a delay of 10 seconds between
copy attempts.

#>
    param
    (
        [Parameter(Mandatory = $true)]
        [string] $SourceFileName,

        [Parameter(Mandatory = $true)]
        [string] $DestinationFileName,

        [Parameter(Mandatory = $true)]
        [int] $TryCount,

        [Parameter(Mandatory = $true)]
        [int] $RetryDelayMilliseconds
    )

    $tryNumber = 1;

    while ($tryNumber -le $TryCount)
    {
        try
        {
            Copy-FileWithRestart -SourceFileName $SourceFileName -DestinationFileName $DestinationFileName;
            return;
        }
        catch [System.Exception]
        {
            $ex = $_.Exception;
            $tryNumber++;

            if ($tryNumber -le $TryCount)
            {
                Start-Sleep -Milliseconds $RetryDelayMilliseconds;
            }
        }
    }

    $message = "Error after trying $TryCount times to copy file. The last error was: " + $ex.Message;
    throw $message;
}

####################################################################################################
## Remove-Files
####################################################################################################
function Remove-Files
{
<#
.SYNOPSIS

Deletes files based on a pattern and a retention period.

.DESCRIPTION

Deletes files 

.PARAMETER DaysToDelete

Specifies the minimum time in days for which files will be deleted.

See the $IgnoreTime parameter comments for additional information.

.PARAMETER MatchString

Specifies a wildcard pattern indicating which files will be deleted.

.PARAMETER TargetDir

Identifies the file path containing the candidate files to be deleted. Note that subdirectories will
be preserved and no files will be deleted from them.

.PARAMETER IgnoreTime

If $true, the current time of day is ignored in determining the minimum age of files. For
example, if today is Thursday at 4:00 pm, and DaysToDelete is 3, matching files before before
MIDNIGHT on Monday will be deleted.

If $false (default), the exact number of days will be used to identify files to be deleted. For
example, if today is Thursday at 4:00 pm, and DaysToDelete is 3, matching files before
Monday at 4:00 pm will be deleted.

.INPUTS

None. You cannot pipe data to Remove-Files.

.OUTPUTS

None.

.EXAMPLE

Remove-Files -DaysToDelete 3 -MatchString "T*.txt" -TargetDir "c:\temp" -IgnoreTime $false;

This command will delete all files matching T*.txt in the c:\temp directory that are three days or older.
If today is Thursday at 4:00 pm, only files older than Monday at 4:00 pm will be deleted.
If the -IgnoreTime parameter were $true, then files older then Monday at midnight would be deleted.

#>
    param
    (
        [Parameter(Mandatory = $true)]
        [int] $DaysToDelete,

        [Parameter(Mandatory = $true)]
        [string] $MatchString,

        [Parameter(Mandatory = $true)]
        [string] $TargetDir,

        [bool] $IgnoreTime = $false
    )

    if ($IgnoreTime -eq $false)
    {
        $cutoffDate = (GetDate).AddMinutes(-$DaysToDelete * 60 * 24);
    }
    else
    {
        $cutoffDate = (Get-Date).Date.AddDays(-$DaysToDelete);
    }

    foreach ($file in Get-ChildItem $TargetDir)
    {
        if (($file.LastWriteTime -lt $cutoffDate) -and ($file.FullName -like $MatchString) -and ($file.FullName -like $MatchString))
        {
            Remove-Item $file.FullName -Force;
        }
    }
}

####################################################################################################
## Copy-TransactionLogs
####################################################################################################
function Copy-TransactionLogs
{
<#
.SYNOPSIS

Copies zero or more files that match a pattern from one directory to another, and then deletes old
versions of the file from the destination that match the same pattern, and optionally logs the 
presence of the newly copied file.

.DESCRIPTION

Copies files matching a transaction log pattern name from one location to another. All files created
in a specified time frame will be copied. Additionally, all files matching the pattern that
are older than a specified threshold will be deleted from the destination directory.

Optionally, a record will be written to a database table indicating that a file has been copied.
This can be used to trigger downstream actions such as verifying the file on a separate system.

.PARAMETER SourcePath

The location where the files will be copied from.

.PARAMETER DestinationPath

The location where the files will be copied to.

.PARAMETER SourceRangeSearchHours

The number of hours used to identify source files. For instance, if this value is 24, only
files created in the last 24 hours will be considered for copy.

The default is 24 hours.

.PARAMETER RetentionHours

The number of hours that files will remain in the destination. Files older than the time indicated
by this value will be removed by this function.

The default is 72 hours (three days).

.PARAMETER CopyRetryCount

In the event that a file copy fails, this indicates the number of times that the function will
attempt to copy the file.

The default is 3 attempts.

.PARAMETER CopyRetryDelaySeconds

In the event that a file copy fails, this indicates the number of seconds that the function will
wait before attempting to copy the file again.

The default is 60 seconds.

.PARAMETER VerifyInstance

An optional value to indicate the name of the instance where a log is maintained of files that are
copied. See the Notes section for additional details.

.PARAMETER VerifyDatabase

An optional value to indicate the name of the database where a log is maintained of files that are
copied.  See the Notes section for additional details.

.PARAMETER SourceSystemName

An optional value that will logged if a log is maintinaed of files are that are copied. This
value is intended to indicate the name of the system that generated the backup file.

.PARAMETER DatabaseList

An array of one or more database names for which files will be copied. This function presumes that
the directory indicated by the -SourcePath contains a subfolder with the database name, and in turn
that the subfolder contains the actual backup files.

.PARAMETER FileExtension

The file extension of the files to be copied.

The default value is "trn".

.PARAMETER ContinueOnFailure

if $true (the default), processing will continue in the event of an error. At the end of processing,
an exception will be raised.

If $false, processing will terminate immediately if an error occurs and an exception will be raised.

.INPUTS

None. You cannot pipe data to Copy-TransactionLogs.

.OUTPUTS

None.

.NOTES

The backup files are presumed to follow the file pattern: *<databaseName>*Log*.<extension>

The specified source directory is presumed to contain a subfolder with the same name as the database
name.

The specified destination directory is presumed to contain a subfolder with the same name as the
database name. If the subfolder does not exist, it will be created when files are copied.

If logging is used, one record per file copied will be written to a table called VerificationBackups.
This table requires the following columns at a minimum (ID is added for convenience, though
it is technically not required):

create table VerificationBackups
(
	ID int identity(1,1) not null primary key,
	BackupFilePath nvarchar(260) not null,
	SourceSystem nvarchar(128) not null,
	BackupType varchar(5) not null,
	FileSize bigint not null,
	CopyStartTime datetime2(7) null,
	CopyEndTime datetime2(7) null,
	CopyRetryAttempts int not null
);

.EXAMPLE

Copy-TransactionLogs -SourcePath "c:\temp" -DestinationPath "d:\temp" -DatabaseList "master";

This command will copy files with a create date in the past 24 hours from c:\temp\master to 
d:\temp\master.

The files much match the pattern *master*Log*.trn

Files in the destination that match the pattern and that are older than 72 hours will be deleted.

The file copy operations will not be logged.

.EXAMPLE

Copy-TransactionLogs -SourcePath "c:\temp" -DestinationPath "d:\temp" -SourceRangeSearchHours 48 -RetentionHours 168 -CopyRetryCount 5 -CopyRetryDelaySeconds 15 -VerifyInstance "sql01\prod" -VerifyDatabase "Admin" -SourceSystemName "sql02\test" -DatabaseList @("master", "model", "msdb") -FileExtension "bak" -ContinueonFailure $false;

This command will copy files with a create date in the past 48 hours from c:\temp\master to 
d:\temp\master for the master database, and so on for the other databases.

The files much match the pattern *master*Log*.bak for the master database, and so on for the 
other databases.

Files for the master, model and msdb database will be processed.

Files in the destination that match the pattern and that are older than 168 hours will be deleted.

In the event of failure, the function will attempt 5 times to copy the time and will
delay 15 seconds between copies.

In the event of an unexpected failure, the function will immediately throw an exception.

The file copy operations will be logged to the VerificationBackups table on the sql01\prod
instance in the Admin database. The value "sql02\test" will be logged as the SourceSystem.

#>
    param
    (
        [Parameter(Mandatory = $true)]
        [string] $SourcePath,

        [Parameter(Mandatory = $true)]
        [string] $DestinationPath,

        [int] $SourceRangeSearchHours = 24,
        [int] $RetentionHours = 72,
        [int] $CopyRetryCount = 3,
        [int] $CopyRetryDelaySeconds = 60,
        [string] $VerifyInstance = $null,
        [string] $VerifyDatabase = $null,
        [string] $SourceSystemName = $null,

        [Parameter(Mandatory = $true)]
        [string[]] $DatabaseList,

        [string] $FileExtension = "trn",
        [bool] $ContinueOnFailure = $true
    )

    if ([String]::IsNullOrEmpty($VerifyInstance) -eq $false)
    {
        $sb = New-Object System.Data.SqlClient.SqlConnectionStringBuilder;
        $sb["Data Source"] = $VerifyInstance;
        $sb["Initial Catalog"] = $VerifyDatabase;
        $sb["Integrated Security"] = $true;
        $connection = New-Object System.Data.SqlClient.SqlConnection($sb.ToString());
        $connection.Open();

        $sql = "insert dbo.VerificationBackups (BackupFilePath, SourceSystem, BackupType, FileSize, CopyStartTime, CopyEndTime, CopyRetryAttempts) values (@BackupFilePath, @SourceSystem, @BackupType, @FileSize, @CopyStartTime, @CopyEndTime, @CopyRetryAttempts);";
    }

    $searchStartTime = (Get-Date).AddHours(-$SourceRangeSearchHours);
    $purgeFromTime = (Get-Date).AddHours(-$RetentionHours);

    foreach($db in $DatabaseList)
    {
        $pattern = "*" + $db + "*Log*." + $FileExtension;
        $searchPath = [System.IO.Path]::Combine($SourcePath, $db);
        $destDbPath = [System.IO.Path]::Combine($DestinationPath, $db);
        $files = Get-ChildItem -Path $searchPath -Filter $pattern | Where-Object { $_.LastWriteTime -gt $searchStartTime };

        foreach ($file in $files)
        {
            $destFile = [System.IO.Path]::Combine($destDbPath, $file.Name);
            if ([System.IO.File]::Exists($destFile) -eq $false -or
                $file.Length -ne (Get-FileSize -FileName $destFile))
            {
                $tryCount = 0;
                $copyStartTime = (Get-Date);
                $stopWatch = [System.Diagnostics.Stopwatch]::StartNew();
                $copySuccessful = $false;

                while ($tryCount -lt $copyRetryCount)
                {
                    try
                    {
                        [void][System.IO.Directory]::CreateDirectory([System.IO.Path]::GetDirectoryName($destFile));
                        [void]$file.CopyTo($destFile, $true);
                        $copySuccessful = $true;
                        break;
                    }
                    catch [System.IO.IOException]
                    {
                        $ex = $_.Exception;
                    }

                    $tryCount++;
                    Start-Sleep $copyRetryDelaySeconds;
                }

                if ($copySuccessful -eq $false)
                {
                    if ($ContinueOnFailure)
                    {
                        if ([System.String]::IsNullOrEmpty($ErrorMessages) -eq $false)
                        {
                            $ErrorMessages += "`r`n";
                        }
                        $ErrorMessages += $ex.Message;
                    }
                    else
                    {
                        throw $ex;
                    }
                }

                $stopWatch.Stop();
                $copyEndTime = $copyStartTime.Add($stopWatch.Elapsed);

                if ([String]::IsNullOrEmpty($VerifyInstance) -eq $false)
                {
                    $command = New-Object System.Data.SqlClient.SqlCommand($sql, $connection);

                    [void]$command.Parameters.Add("@BackupFilePath", $destFile);
                    [void]$command.Parameters.Add("@SourceSystem", $sourceSystemName);
                    [void]$command.Parameters.Add("@BackupType", "Log");
                    [void]$command.Parameters.Add("@FileSize", $file.Length);

                    $startParameter = New-Object System.Data.SqlClient.SqlParameter("@CopyStartTime", $copyStartTime);
                    $startParameter.SqlDbType = [System.Data.SqlDbType]::Datetime2;
                    [void]$command.Parameters.Add($startParameter);

                    $endParameter = New-Object System.Data.SqlClient.SqlParameter("@CopyEndTime", $copyEndTime);
                    $endParameter.SqlDbType = [System.Data.SqlDbType]::Datetime2;
                    [void]$command.Parameters.Add($endParameter);

                    [void]$command.Parameters.Add("@CopyRetryAttempts", $tryCount);
                    $command.CommandTimeout = 300;

                    [void]$command.ExecuteNonQuery();
                    $command.Dispose();
                }
            }
        }

        if ([System.IO.Directory]::Exists($destDbPath))
        {
            Get-ChildItem -Path $destDbPath -Filter $pattern | Where-Object { $_.LastWriteTime -lt $purgeFromTime } | ForEach-Object { $_.Delete() };
        }
    }

    if ([String]::IsNullOrEmpty($VerifyInstance) -eq $false)
    {
        $connection.Close();
    }

    if ($ContinueOnFailure -and [System.String]::IsNullOrEmpty($ErrorMessages) -eq $false)
    {
        throw $ErrorMessages;
    }
}

####################################################################################################
## Copy-MostRecentBackups
####################################################################################################
function Copy-MostRecentBackups
{
<#
.SYNOPSIS

Copies zero or one file that match a pattern from one directory to another, and then deletes old
versions of the file from the destination that match the same pattern, and optionally logs the 
presence of the newly copied file.

.DESCRIPTION

Copies files matching a full backup pattern name from one location to another. Only the most recent
file will be copied. Additionally, all files matching the pattern that are older than a specified 
threshold will be deleted from the destination directory.

Optionally, a record will be written to a database table indicating that a file has been copied.
This can be used to trigger downstream actions such as verifying the file on a separate system.

.PARAMETER SourcePath

The location where the files will be copied from.

.PARAMETER DestinationPath

The location where the files will be copied to.

.PARAMETER RetentionDays

The number of days that files will remain in the destination. Files older than the time indicated
by this value will be removed by this function.

The default is 3 days.

.PARAMETER RetentionDaysIgnoreTime

If $true, the current time of day is ignored in determining the minimum age of files. For
example, if today is Thursday at 4:00 pm, and RetentionDays is 3, matching files before before
MIDNIGHT on Monday will be deleted.

If $false (default), the exact number of days will be used to identify files to be deleted. For
example, if today is Thursday at 4:00 pm, and RetentionDays is 3, matching files before
Monday at 4:00 pm will be deleted.

The default is 3 days.

.PARAMETER CopyRetryCount

In the event that a file copy fails, this indicates the number of times that the function will
attempt to copy the file.

The default is 3 attempts.

.PARAMETER CopyRetryDelaySeconds

In the event that a file copy fails, this indicates the number of seconds that the function will
wait before attempting to copy the file again.

The default is 60 seconds.

.PARAMETER VerifyInstance

An optional value to indicate the name of the instance where a log is maintained of files that are
copied. See the Notes section for additional details.

.PARAMETER VerifyDatabase

An optional value to indicate the name of the database where a log is maintained of files that are
copied.  See the Notes section for additional details.

.PARAMETER SourceSystemName

An optional value that will logged if a log is maintinaed of files are that are copied. This
value is intended to indicate the name of the system that generated the backup file.

.PARAMETER DatabaseList

An array of one or more database names for which files will be copied. This function presumes that
the directory indicated by the -SourcePath contains a subfolder with the database name, and in turn
that the subfolder contains the actual backup files.

.PARAMETER BackupType

A value indicating the type of the backup file. This value is only used for pattern matching.
Typical values are "Full" or "Diff".

.PARAMETER FileExtension

The file extension of the files to be copied.

The default value is "bak".

.PARAMETER ContinueOnFailure

if $true (the default), processing will continue in the event of an error. At the end of processing,
an exception will be raised.

If $false, processing will terminate immediately if an error occurs and an exception will be raised.

.INPUTS

None. You cannot pipe data to Copy-MostRecentBackups.

.OUTPUTS

None.

.NOTES

The backup files are presumed to follow the file pattern: *<databaseName>*<backupType>*.<extension>

The specified source directory is presumed to contain a subfolder with the same name as the database
name.

The specified destination directory is presumed to contain a subfolder with the same name as the
database name. If the subfolder does not exist, it will be created when files are copied.

If logging is used, one record per file copied will be written to a table called VerificationBackups.
This table requires the following columns at a minimum (ID is added for convenience, though
it is technically not required):

create table VerificationBackups
(
	ID int identity(1,1) not null primary key,
	BackupFilePath nvarchar(260) not null,
	SourceSystem nvarchar(128) not null,
	BackupType varchar(5) not null,
	FileSize bigint not null,
	CopyStartTime datetime2(7) null,
	CopyEndTime datetime2(7) null,
	CopyRetryAttempts int not null
);

.EXAMPLE

Copy-MostRecentBackups -SourcePath "c:\temp" -DestinationPath "d:\temp" -DatabaseList "master" -BackupType "Full";

This command will copy the most recent file in c:\temp\master to d:\temp\master.

The files much match the pattern *master*Full*.bak

Files in the destination that match the pattern and that are older than 3 days will be deleted.

The file copy operations will not be logged.

.EXAMPLE

Copy-MostRecentBackups -SourcePath "c:\temp" -DestinationPath "d:\temp" -SourceRangeSearchHours 48 -RetentionDays 7 -CopyRetryCount 5 -CopyRetryDelaySeconds 15 -VerifyInstance "sql01\prod" -VerifyDatabase "Admin" -SourceSystemName "sql02\test" -DatabaseList @("master", "model", "msdb") -BackupType "Diff" -FileExtension "bak" -ContinueonFailure $false;

This command will copy the most recent file in c:\temp\master to d:\temp\master. for the master 
database, and so on for the other databases.

The files much match the pattern *master*Diff*.bak for the master database, and so on for the 
other databases.

Files for the master, model and msdb database will be processed.

Files in the destination that match the pattern and that are older than 7 days will be deleted.

In the event of failure, the function will attempt 5 times to copy the time and will
delay 15 seconds between copies.

In the event of an unexpected failure, the function will immediately throw an exception.

The file copy operations will be logged to the VerificationBackups table on the sql01\prod
instance in the Admin database. The value "sql02\test" will be logged as the SourceSystem.

#>
    param
    (
        [Parameter(Mandatory = $true)]
        [string] $SourcePath,

        [Parameter(Mandatory = $true)]
        [string] $DestinationPath,

        [int] $RetentionDays = 3,
        [bool] $RetentionIgnoreTime = $false,
        [int] $CopyRetryCount = 3,
        [int] $CopyRetryDelaySeconds = 60,
        [string] $VerifyInstance = $null,
        [string] $VerifyDatabase = $null,
        [string] $SourceSystemName = $null,

        [Parameter(Mandatory = $true)]
        [string[]] $DatabaseList,

        [Parameter(Mandatory = $true)]
        [string] $BackupType,

        [string] $FileExtension = "bak",
        [bool] $ContinueOnFailure = $true
    )

    $ErrorMessages = "";

    if ([String]::IsNullOrEmpty($VerifyInstance) -eq $false)
    {
        $sb = New-Object System.Data.SqlClient.SqlConnectionStringBuilder;
        $sb["Data Source"] = $VerifyInstance;
        $sb["Initial Catalog"] = $VerifyDatabase;
        $sb["Integrated Security"] = $true;
        $connection = New-Object System.Data.SqlClient.SqlConnection($sb.ToString());
        $connection.Open();

        $sql = "insert dbo.VerificationBackups (BackupFilePath, SourceSystem, BackupType, FileSize, CopyStartTime, CopyEndTime, CopyRetryAttempts) values (@BackupFilePath, @SourceSystem, @BackupType, @FileSize, @CopyStartTime, @CopyEndTime, @CopyRetryAttempts);";
    }

    if ($RetentionIgnoreTime -eq $false)
    {
        $purgeFromTime = (Get-Date).AddDays(-$RetentionDays);
    }
    else
    {
        $purgeFromTime = (Get-Date).Date.AddDays(-$RetentionDays);
    }

    foreach($db in $DatabaseList)
    {
        $pattern = "*" + $db + "*$BackupType*." + $FileExtension;
        $searchPath = [System.IO.Path]::Combine($SourcePath, $db);
        $destDbPath = [System.IO.Path]::Combine($DestinationPath, $db);

        $file = Get-ChildItem -Path $searchPath -Filter $pattern | Where-Object { $_.LastWriteTime -ge  $purgeFromTime} | Sort-Object -Property LastWriteTime -Descending | Select-Object -First 1;

        if ($null -ne $file)
        {
            $destFile = [System.IO.Path]::Combine($destDbPath, $file.Name);
            if ([System.IO.File]::Exists($destFile) -eq $false -or
                $file.Length -ne (Get-FileSize -FileName $destFile))
            {
                $tryCount = 0;
                $copyStartTime = (Get-Date);
                $stopWatch = [System.Diagnostics.Stopwatch]::StartNew();
                $copySuccessful = $false;

                while ($tryCount -lt $copyRetryCount)
                {
                    try
                    {
                        [void][System.IO.Directory]::CreateDirectory([System.IO.Path]::GetDirectoryName($destFile));
                        [void]$file.CopyTo($destFile, $true);
                        $copySuccessful = $true;
                        break;
                    }
                    catch [System.IO.IOException]
                    {
                        $ex = $_.Exception;
                    }

                    $tryCount++;
                    Start-Sleep $copyRetryDelaySeconds;
                }

                $stopWatch.Stop();
                $copyEndTime = $copyStartTime.Add($stopWatch.Elapsed);

                if ($copySuccessful -eq $false)
                {
                    if ($ContinueOnFailure)
                    {
                        if ([System.String]::IsNullOrEmpty($ErrorMessages) -eq $false)
                        {
                            $ErrorMessages += "`r`n";
                        }
                        $ErrorMessages += $ex.Message;
                    }
                    else
                    {
                        throw $ex;
                    }
                }
                else
                {
                    if ([String]::IsNullOrEmpty($VerifyInstance) -eq $false)
                    {
                        $command = New-Object System.Data.SqlClient.SqlCommand($sql, $connection);

                        [void]$command.Parameters.Add("@BackupFilePath", $destFile);
                        [void]$command.Parameters.Add("@SourceSystem", $sourceSystemName);
                        [void]$command.Parameters.Add("@BackupType", "Log");
                        [void]$command.Parameters.Add("@FileSize", $file.Length);

                        $startParameter = New-Object System.Data.SqlClient.SqlParameter("@CopyStartTime", $copyStartTime);
                        $startParameter.SqlDbType = [System.Data.SqlDbType]::Datetime2;
                        [void]$command.Parameters.Add($startParameter);

                        $endParameter = New-Object System.Data.SqlClient.SqlParameter("@CopyEndTime", $copyEndTime);
                        $endParameter.SqlDbType = [System.Data.SqlDbType]::Datetime2;
                        [void]$command.Parameters.Add($endParameter);

                        [void]$command.Parameters.Add("@CopyRetryAttempts", $tryCount);
                        $command.CommandTimeout = 300;

                        [void]$command.ExecuteNonQuery();
                        $command.Dispose();
                    }
                }
            }
        }

        if ([System.IO.Directory]::Exists($destDbPath))
        {
            Get-ChildItem -Path $destDbPath -Filter $pattern | Where-Object { $_.LastWriteTime -lt $purgeFromTime } | ForEach-Object { $_.Delete() };
        }
    }

    if ([String]::IsNullOrEmpty($VerifyInstance) -eq $false)
    {
        $connection.Close();
    }

    if ($ContinueOnFailure -and [System.String]::IsNullOrEmpty($ErrorMessages) -eq $false)
    {
        throw $ErrorMessages;
    }
}

####################################################################################################
## Copy-MostRecentFullBackups
####################################################################################################
function Copy-MostRecentFullBackups
{
<#
.SYNOPSIS

Copies zero or one file that match a pattern from one directory to another, and then deletes old
versions of the file from the destination that match the same pattern, and optionally logs the 
presence of the newly copied file.

.DESCRIPTION

Copies files matching a full backup pattern name from one location to another. Only the most recent
file will be copied. Additionally, all files matching the pattern that are older than a specified 
threshold will be deleted from the destination directory.

Optionally, a record will be written to a database table indicating that a file has been copied.
This can be used to trigger downstream actions such as verifying the file on a separate system.

.PARAMETER SourcePath

The location where the files will be copied from.

.PARAMETER DestinationPath

The location where the files will be copied to.

.PARAMETER RetentionDays

The number of days that files will remain in the destination. Files older than the time indicated
by this value will be removed by this function.

The default is 3 days.

.PARAMETER RetentionDaysIgnoreTime

If $true, the current time of day is ignored in determining the minimum age of files. For
example, if today is Thursday at 4:00 pm, and RetentionDays is 3, matching files before before
MIDNIGHT on Monday will be deleted.

If $false (default), the exact number of days will be used to identify files to be deleted. For
example, if today is Thursday at 4:00 pm, and RetentionDays is 3, matching files before
Monday at 4:00 pm will be deleted.

The default is 3 days.

.PARAMETER CopyRetryCount

In the event that a file copy fails, this indicates the number of times that the function will
attempt to copy the file.

The default is 3 attempts.

.PARAMETER CopyRetryDelaySeconds

In the event that a file copy fails, this indicates the number of seconds that the function will
wait before attempting to copy the file again.

The default is 60 seconds.

.PARAMETER VerifyInstance

An optional value to indicate the name of the instance where a log is maintained of files that are
copied. See the Notes section for additional details.

.PARAMETER VerifyDatabase

An optional value to indicate the name of the database where a log is maintained of files that are
copied.  See the Notes section for additional details.

.PARAMETER SourceSystemName

An optional value that will logged if a log is maintinaed of files are that are copied. This
value is intended to indicate the name of the system that generated the backup file.

.PARAMETER DatabaseList

An array of one or more database names for which files will be copied. This function presumes that
the directory indicated by the -SourcePath contains a subfolder with the database name, and in turn
that the subfolder contains the actual backup files.

.PARAMETER FileExtension

The file extension of the files to be copied.

The default value is "bak".

.PARAMETER ContinueOnFailure

if $true (the default), processing will continue in the event of an error. At the end of processing,
an exception will be raised.

If $false, processing will terminate immediately if an error occurs and an exception will be raised.

.INPUTS

None. You cannot pipe data to Copy-MostRecentFullBackups.

.OUTPUTS

None.

.NOTES

The backup files are presumed to follow the file pattern: *<databaseNameba>*Full*.<extension>

The specified source directory is presumed to contain a subfolder with the same name as the database
name.

The specified destination directory is presumed to contain a subfolder with the same name as the
database name. If the subfolder does not exist, it will be created when files are copied.

If logging is used, one record per file copied will be written to a table called VerificationBackups.
This table requires the following columns at a minimum (ID is added for convenience, though
it is technically not required):

create table VerificationBackups
(
	ID int identity(1,1) not null primary key,
	BackupFilePath nvarchar(260) not null,
	SourceSystem nvarchar(128) not null,
	BackupType varchar(5) not null,
	FileSize bigint not null,
	CopyStartTime datetime2(7) null,
	CopyEndTime datetime2(7) null,
	CopyRetryAttempts int not null
);

.EXAMPLE

Copy-MostRecentFullBackups -SourcePath "c:\temp" -DestinationPath "d:\temp" -DatabaseList "master";

This command will copy the most recent file in c:\temp\master to d:\temp\master.

The files much match the pattern *master*Full*.bak

Files in the destination that match the pattern and that are older than 3 days will be deleted.

The file copy operations will not be logged.

.EXAMPLE

Copy-MostRecentFullBackups -SourcePath "c:\temp" -DestinationPath "d:\temp" -SourceRangeSearchHours 48 -RetentionDays 7 -CopyRetryCount 5 -CopyRetryDelaySeconds 15 -VerifyInstance "sql01\prod" -VerifyDatabase "Admin" -SourceSystemName "sql02\test" -DatabaseList @("master", "model", "msdb") -FileExtension "bak" -ContinueonFailure $false;

This command will copy the most recent file in c:\temp\master to d:\temp\master. for the master 
database, and so on for the other databases.

The files much match the pattern *master*Full*.bak for the master database, and so on for the 
other databases.

Files for the master, model and msdb database will be processed.

Files in the destination that match the pattern and that are older than 7 days will be deleted.

In the event of failure, the function will attempt 5 times to copy the time and will
delay 15 seconds between copies.

In the event of an unexpected failure, the function will immediately throw an exception.

The file copy operations will be logged to the VerificationBackups table on the sql01\prod
instance in the Admin database. The value "sql02\test" will be logged as the SourceSystem.

#>
    param
    (
        [string] $SourcePath,
        [string] $DestinationPath,
        [int] $RetentionDays = 3,
        [bool] $RetentionIgnoreTime = $false,
        [int] $CopyRetryCount = 3,
        [int] $CopyRetryDelaySeconds = 60,
        [string] $VerifyInstance = $null,
        [string] $VerifyDatabase = $null,
        [string] $SourceSystemName = $null,
        [string[]] $DatabaseList,
        [string] $FileExtension = "bak",
        [bool] $ContinueOnFailure = $true
    )

    Copy-MostRecentBackups `
        -SourcePath $SourcePath `
        -DestinationPath $DestinationPath `
        -RetentionDays $RetentionDays `
        -RetentionIgnoreTime $RetentionIgnoreTime `
        -CopyRetryCount $CopyRetryCount `
        -CopyRetryDelaySeconds $CopyRetryDelaySeconds `
        -VerifyInstance $VerifyInstance `
        -VerifyDatabase $VerifyDatabase `
        -SourceSystemName $SourceSystemName `
        -DatabaseList $DatabaseList `
        -BackupType "Full" `
        -FileExtension $FileExtension `
        -ContinueOnFailure $ContinueOnFailure;
}

####################################################################################################
## Copy-MostRecentDiffBackups
####################################################################################################
function Copy-MostRecentDiffBackups
{
<#
.SYNOPSIS

Copies zero or one file that match a pattern from one directory to another, and then deletes old
versions of the file from the destination that match the same pattern, and optionally logs the 
presence of the newly copied file.

.DESCRIPTION

Copies files matching a full backup pattern name from one location to another. Only the most recent
file will be copied. Additionally, all files matching the pattern that are older than a specified 
threshold will be deleted from the destination directory.

Optionally, a record will be written to a database table indicating that a file has been copied.
This can be used to trigger downstream actions such as verifying the file on a separate system.

.PARAMETER SourcePath

The location where the files will be copied from.

.PARAMETER DestinationPath

The location where the files will be copied to.

.PARAMETER RetentionDays

The number of days that files will remain in the destination. Files older than the time indicated
by this value will be removed by this function.

The default is 3 days.

.PARAMETER RetentionDaysIgnoreTime

If $true, the current time of day is ignored in determining the minimum age of files. For
example, if today is Thursday at 4:00 pm, and RetentionDays is 3, matching files before before
MIDNIGHT on Monday will be deleted.

If $false (default), the exact number of days will be used to identify files to be deleted. For
example, if today is Thursday at 4:00 pm, and RetentionDays is 3, matching files before
Monday at 4:00 pm will be deleted.

The default is 3 days.

.PARAMETER CopyRetryCount

In the event that a file copy fails, this indicates the number of times that the function will
attempt to copy the file.

The default is 3 attempts.

.PARAMETER CopyRetryDelaySeconds

In the event that a file copy fails, this indicates the number of seconds that the function will
wait before attempting to copy the file again.

The default is 60 seconds.

.PARAMETER VerifyInstance

An optional value to indicate the name of the instance where a log is maintained of files that are
copied. See the Notes section for additional details.

.PARAMETER VerifyDatabase

An optional value to indicate the name of the database where a log is maintained of files that are
copied.  See the Notes section for additional details.

.PARAMETER SourceSystemName

An optional value that will logged if a log is maintinaed of files are that are copied. This
value is intended to indicate the name of the system that generated the backup file.

.PARAMETER DatabaseList

An array of one or more database names for which files will be copied. This function presumes that
the directory indicated by the -SourcePath contains a subfolder with the database name, and in turn
that the subfolder contains the actual backup files.

.PARAMETER FileExtension

The file extension of the files to be copied.

The default value is "bak".

.PARAMETER ContinueOnFailure

if $true (the default), processing will continue in the event of an error. At the end of processing,
an exception will be raised.

If $false, processing will terminate immediately if an error occurs and an exception will be raised.

.INPUTS

None. You cannot pipe data to Copy-MostRecentBackups.

.OUTPUTS

None.

.NOTES

The backup files are presumed to follow the file pattern: *<databaseNameba>*Full*.<extension>

The specified source directory is presumed to contain a subfolder with the same name as the database
name.

The specified destination directory is presumed to contain a subfolder with the same name as the
database name. If the subfolder does not exist, it will be created when files are copied.

If logging is used, one record per file copied will be written to a table called VerificationBackups.
This table requires the following columns at a minimum (ID is added for convenience, though
it is technically not required):

create table VerificationBackups
(
	ID int identity(1,1) not null primary key,
	BackupFilePath nvarchar(260) not null,
	SourceSystem nvarchar(128) not null,
	BackupType varchar(5) not null,
	FileSize bigint not null,
	CopyStartTime datetime2(7) null,
	CopyEndTime datetime2(7) null,
	CopyRetryAttempts int not null
);

.EXAMPLE

Copy-MostRecentDiffBackups -SourcePath "c:\temp" -DestinationPath "d:\temp" -DatabaseList "master";

This command will copy the most recent file in c:\temp\master to d:\temp\master.

The files much match the pattern *master*Diff*.bak

Files in the destination that match the pattern and that are older than 3 days will be deleted.

The file copy operations will not be logged.

.EXAMPLE

Copy-MostRecentDiffBackups -SourcePath "c:\temp" -DestinationPath "d:\temp" -SourceRangeSearchHours 48 -RetentionDays 7 -CopyRetryCount 5 -CopyRetryDelaySeconds 15 -VerifyInstance "sql01\prod" -VerifyDatabase "Admin" -SourceSystemName "sql02\test" -DatabaseList @("master", "model", "msdb") -FileExtension "bak" -ContinueonFailure $false;

This command will copy the most recent file in c:\temp\master to d:\temp\master. for the master 
database, and so on for the other databases.

The files much match the pattern *master*Diff*.bak for the master database, and so on for the 
other databases.

Files for the master, model and msdb database will be processed.

Files in the destination that match the pattern and that are older than 7 days will be deleted.

In the event of failure, the function will attempt 5 times to copy the time and will
delay 15 seconds between copies.

In the event of an unexpected failure, the function will immediately throw an exception.

The file copy operations will be logged to the VerificationBackups table on the sql01\prod
instance in the Admin database. The value "sql02\test" will be logged as the SourceSystem.

#>
    param
    (
        [string] $SourcePath,
        [string] $DestinationPath,
        [int] $RetentionDays = 3,
        [bool] $RetentionIgnoreTime = $false,
        [int] $CopyRetryCount = 3,
        [int] $CopyRetryDelaySeconds = 60,
        [string] $VerifyInstance = $null,
        [string] $VerifyDatabase = $null,
        [string] $SourceSystemName = $null,
        [string[]] $DatabaseList,
        [string] $FileExtension = "bak",
        [bool] $ContinueOnFailure = $true
    )

    Copy-MostRecentBackups `
        -SourcePath $SourcePath `
        -DestinationPath $DestinationPath `
        -RetentionDays $RetentionDays `
        -RetentionIgnoreTime $RetentionIgnoreTime `
        -CopyRetryCount $CopyRetryCount `
        -CopyRetryDelaySeconds $CopyRetryDelaySeconds `
        -VerifyInstance $VerifyInstance `
        -VerifyDatabase $VerifyDatabase `
        -SourceSystemName $SourceSystemName `
        -DatabaseList $DatabaseList `
        -BackupType "Diff" `
        -FileExtension $FileExtension `
        -ContinueOnFailure $ContinueOnFailure;
}

#####################################################################################################
## Get-RestoreInfoNative function
#####################################################################################################
function Get-RestoreInfoNative
{
    param
    (
        [Parameter(Mandatory = $true)]
        [string] $BackupFileName,

        [Parameter(Mandatory = $true)]
        [string] $RestoreServerName
    )

    Add-CustomTypesInternal;

    $connectionBuilder = New-Object System.Data.SqlClient.SqlConnectionStringBuilder;
    $connectionBuilder["Data Source"] = $RestoreServerName;
    $connectionBuilder["Integrated Security"] = $true;

    #$sql = "restore headeronly from disk = '" + $BackupFileName + "';";
    $sql = "restore headeronly from disk = @BackupFileName;";

    $table = New-Object System.Data.DataTable;
    try
    {
        try
        {
            $sqlConnection = New-Object System.Data.SqlClient.SqlConnection($connectionBuilder.ToString());
            $sqlConnection.Open();

            try
            {
                $sqlCommand = New-Object System.Data.SqlClient.SqlCommand($sql, $sqlConnection);
                $sqlCommand.CommandTimeout = 3600;
                [void]$sqlCommand.Parameters.Add("@BackupFileName", $BackupFileName);

                try
                {
                    $adapter = New-Object System.Data.SqlClient.SqlDataAdapter($sqlCommand);
                    [void]$adapter.Fill($table);
                    if ($table.Rows.Count -eq 0)
                    {
                        return $null;
                    }
                }
                finally
                {
                    if ($null -ne $adapter)
                    {
                        $adapter.Dispose();
                    }
                }
            }
            finally
            {
                if ($null -ne $sqlCommand)
                {
                    $sqlCommand.Dispose();
                }
            }
        }
        catch [System.Data.SqlClient.SqlException]
        {
            $errMsg = "Error executing SQL: $_.Exception.Message (SQL = $sql)";
            $ex = New-Object System.Exception($errMsg, $_.Exception);
            throw $ex;
        }
        finally
        {
            if ($null -ne $sqlConnection)
            {
                $sqlConnection.Close();
            }
        }

        if ($table.Rows.Count -eq 0)
        {
            throw "No backups found within backup file $BackupFileName";
        }

        $row = $table.Rows[0];

        $result = New-Object Sql.RestoreInfo;

        $result.BackupFileName = $BackupFileName;
        $result.FirstLSN = $row["FirstLSN"];
        $result.LastLSN = $row["LastLSN"];
        $result.CheckpointLSN = $row["CheckpointLSN"];
        $result.DatabaseBackupLSN = $row["DatabaseBackupLSN"];
        $result.BackupSize = $row["BackupSize"];
        $result.BackupStartDate = $row["BackupStartDate"];
        $result.BackupFinishDate = $row["BackupFinishDate"];
        $result.BackupType = "";

        $type = $row["BackupType"];
        if ($type -eq 1) { $result.BackupType = "Full"; }
        if ($type -eq 2) { $result.BackupType = "Log"; }
        if ($type -eq 4) { $result.BackupType = "Filegroup"; }
        if ($type -eq 5) { $result.BackupType = "Differential"; }
        if ($type -eq 6) { $result.BackupType = "Differential Filegroup"; }
        if ($type -eq 7) { $result.BackupType = "Partial"; }
        if ($type -eq 8) { $result.BackupType = "Differential Partial"; }

        $result.ServerName = $row["ServerName"];
        $result.DatabaseName = $row["DatabaseName"];
        $result.Compressed = $row["Compressed"];
        $result.Position = $row["Position"];
        $result.CompatibilityLevel = $row["CompatibilityLevel"];
        $result.Encryption = $row["Encryption"];
        $result.IsCopyOnly = $row["IsCopyOnly"];

        return $result;
    }
    finally
    {
        if ($null -ne $table)
        {
            $table.Dispose();
        }
    }
}

#####################################################################################################
## Get-RestoreInfo function
#####################################################################################################
function Get-RestoreInfo
{
    param
    (
        [Parameter(Mandatory = $true)]
        [string] $BackupFileName,

        [Parameter(Mandatory = $true)]
        [string] $RestoreServerName,

        [string] $BackupApplication = "Native"
    )

    if ($BackupApplication -eq "Native")
    {
        Get-RestoreInfoNative -BackupFileName $BackupFileName -RestoreServerName $RestoreServerName;
    }
    else
    {
        throw "Unknown value for BackupApplication. Valid values are `"Native`".";
    }
}

#####################################################################################################
## Get-FileListOnlyInfo function
#####################################################################################################
function Get-FileListOnlyInfo
{
    param
    (
        [Parameter(Mandatory = $true)]
        [string] $BackupFileName,

        [Parameter(Mandatory = $true)]
        [string] $RestoreServerName
    )

    Add-CustomTypesInternal;

    $connectionBuilder = New-Object System.Data.SqlClient.SqlConnectionStringBuilder;
    $connectionBuilder["Data Source"] = $RestoreServerName;
    $connectionBuilder["Integrated Security"] = $true;

    $sql = "restore filelistonly from disk = @backupFileName;";

    try
    {
        $sqlConnection = New-Object System.Data.SqlClient.SqlConnection($connectionBuilder.ToString());
        $sqlConnection.Open();

        $fileParameter = New-Object System.Data.SqlClient.SqlParameter("BackupFileName", $BackupFileName);
        $table = (Get-DataTableFromQuery -SqlConnection $sqlConnection -Sql $sql -Parameters $fileParameter);
    }
    catch [System.Data.SqlClient.SqlException]
    {
        $errMsg = "Error executing SQL: $_.Exception.Message (SQL = $sql)";
        $ex = New-Object System.Exception($errMsg, $_.Exception);
        throw $ex;
    }
    finally
    {
        if ($null -ne $sqlConnection)
        {
            $sqlConnection.Close();
        }
    }

    if ($table.Rows.Count -eq 0)
    {
        throw "No files found with backup file $BackupFileName";
    }

    $fileListOnlyList = New-Object Sql.FileListOnlyInfoList;
    foreach($row in $table.Rows)
    {
        $item = New-Object Sql.FileListOnlyInfo;

        $item.BackupFileName = $BackupFileName;
        $item.LogicalName = $row["LogicalName"];
        $item.PhysicalName = $row["PhysicalName"];

        $backupFileType = $row["Type"];
        if ($backupFileType -eq "L") { $item.BackupFileType = "Log"; }
        if ($backupFileType -eq "D") { $item.BackupFileType = "Full"; }
        if ($backupFileType -eq "F") { $item.BackupFileType = "FullTextCatalog"; }

        $item.FileGroupName = $row["FileGroupName"];
        $item.Size = $row["Size"];
        $item.MaxSize = $row["MaxSize"];
        $item.FileID = $row["FileID"];
        $item.CreateLsn = $row["CreateLsn"];
        $item.DropLsn = $row["DropLsn"];
        $item.UniqueId = $row["UniqueId"];
        $item.ReadOnlyLsn = $row["ReadOnlyLsn"];
        $item.ReadWriteLsn = $row["ReadWriteLsn"];
        $item.BackupSizeInBytes = $row["BackupSizeInBytes"];
        $item.SourceBlockSize = $row["SourceBlockSize"];
        $item.FileGroupId = $row["FileGroupId"];
        $item.DifferentialBaseLsn = $row["DifferentialBaseLsn"];
        $item.DifferentialBaseGuid = $row["DifferentialBaseGuid"];
        $item.IsReadOnly = $row["IsReadOnly"];
        $item.IsPresent = $row["IsPresent"];

        $tdeThumbprint = $row["TdeThumbprint"];
        if ($tdeThumbprint -ne [System.DBNull]::Value)
        {
            $item.TdeThumbprint = $tdeThumbprint;
        }

        $fileListOnlyList.Add($item);
    }

    return $fileListOnlyList;
}

####################################################################################################
## Restore-DatabaseInternal
####################################################################################################
function Restore-DatabaseInternal
{
    param
    (
        [Parameter(Mandatory = $true)]
        [string] $InstanceName,

        [Parameter(Mandatory = $true)]
        [string] $DatabaseName,

        [string] $BackupFileName,

        [Parameter(Mandatory = $true)]
        [string] $BackupType,

        [string] $MdfPathName = $null,
        [string] $LdfPathName = $null,
        [Nullable[DateTime]] $RestorePoint = $null,
        [int] $BackupTimeoutSeconds = 600,
        [bool] $SetToSimple = $false,
        [bool] $OutputScriptOnly = $false
    )

    $ErrorMessages = "";
    $restoreParameterValue = [System.DBNull]::Value;
    $restoreScriptSql = "";

    if ($BackupType -ne "Recovery" -and [System.String]::IsNullOrEmpty($BackupFileName) -eq $true)
    {
        throw "BackupFileName parameter is required unless BackupType is 'Recovery'";
    }

    if ($null -ne $MdfPathName -and $null -eq $LdfPathName)
    {
        $LdfPathName = $MdfPathName;
    }

    $fromClause = "from disk = '$backupFileName'";
    $differentialClause = "";
    $stopAtClause = "";
    $recoveryClause = "";
    $replaceClause = "";
    $moveClause = "";

    if ($BackupType -eq "Full")
    {
        $commandType = "database";
        $recoveryClause = "norecovery";
        $replaceClause = ", replace";

        if ([System.String]::IsNullOrEmpty($MdfPathName) -eq $false)
        {
            $fileListOnlyList = Get-FileListOnlyInfo -BackupFileName $BackupFileName -RestoreServerName $InstanceName;
            foreach($fileListItem in $fileListOnlyList)
            {
                if ($fileListItem.BackupFileType -eq "Full")
                {
                    $newFileName = [System.IO.Path]::Combine($MdfPathName, [System.IO.Path]::GetFileName($fileListItem.PhysicalName));
                    $logicalName = $fileListItem.LogicalName;
                    $moveClause += ", move '$logicalName' to '$newFileName'";
                }
                elseif ($fileListItem.BackupFileType -eq "Log")
                {
                    $newFileName = [System.IO.Path]::Combine($LdfPathName, [System.IO.Path]::GetFileName($fileListItem.PhysicalName));
                    $logicalName = $fileListItem.LogicalName;
                    $moveClause += ", move '$logicalName' to '$newFileName'";
                }
            }
        }
    }
    elseif ($BackupType -eq "Diff" -or $BackupType -eq "Differential")
    {
        $commandType = "database";
        $differentialClause = ", differential";
        $recoveryClause = "norecovery";
    }
    elseif ($BackupType -eq "Log")
    {
        $commandType = "log";
        if ($null -ne $RestorePoint)
        {
            $stopAtClause = ", stopat = @restorePoint";
            $restoreParameterValue = $restorePoint;
        }
        $recoveryClause = "norecovery";
    }
    elseif ($BackupType -eq "Recovery")
    {
        $commandType = "database";
        $recoveryClause = "recovery";
        $fromClause = "";
    }
    else
    {
        throw "Invalid BackupType specified: $BackupType";
    }

    $restorePointParameter = New-Object System.Data.SqlClient.SqlParameter("@restorePoint", $restoreParameterValue);

    $sb = New-Object System.Data.SqlClient.SqlConnectionStringBuilder;
    $sb["Data Source"] = $InstanceName;
    $sb["Integrated Security"] = $true;

    try
    {
        $connection = New-Object System.Data.SqlClient.SqlConnection($sb.ToString());
        $connection.Open();

        $offlineSql = "if exists (select * from sys.databases where name = '$DatabaseName' and state_desc = 'ONLINE') alter database [$DatabaseName] set offline with rollback immediate;";
        $restoreSql = "restore $commandType [$DatabaseName] $fromClause with $recoveryClause $replaceClause $differentialClause $stopAtClause $moveClause;";
        $simpleSql = "alter database [$DatabaseName] set recovery simple;";

        try
        {
            if ($BackupType -eq "Full")
            {
                if ($OutputScriptOnly -eq $false)
                {
                    $command = New-Object System.Data.SqlClient.SqlCommand($offlineSql, $connection);
                    $command.CommandTimeout = $BackupTimeoutSeconds;
                    [void]$command.ExecuteNonQuery();
                }
                else
                {
                    $restoreScriptSql += $offlineSql + "`r`n";
                }
            }

            if ($OutputScriptOnly -eq $false)
            {
                $command = New-Object System.Data.SqlClient.SqlCommand($restoreSql, $connection);
                [void]$command.Parameters.Add($restorePointParameter);
                $command.CommandTimeout = $BackupTimeoutSeconds;
                [void]$command.ExecuteNonQuery();
            }
            else
            {
                if ($restoreParameterValue -ne [System.DBNull]::Value)
                {
                    $restorePointString = $restoreParameterValue.ToString("yyyy-MM-dd HH:mm:ss.fff");
                    $restoreScriptSql += "declare @restorePoint datetime = '$restorePointString';`r`n";
                }
                $restoreScriptSql += $restoreSql + "`r`n";
            }

            if ($BackupType -eq "Recovery" -and $SetToSimple -eq $true)
            {
                if ($OutputScriptOnly -eq $false)
                {
                    $command = New-Object System.Data.SqlClient.SqlCommand($simpleSql, $connection);
                    $command.CommandTimeout = $BackupTimeoutSeconds;
                    [void]$command.ExecuteNonQuery();
                }
                else
                {
                    $restoreScriptSql += $simpleSql + "`r`n";
                }
            }
        }
        catch [Exception]
        {
            if ([System.String]::IsNullOrEmpty($ErrorMessages) -eq $false)
            {
                $ErrorMessages += "`r`n";
            }
            $ErrorMessages += $_.Exception.Message;
        }
        finally
        {
            if ($null -ne $command)
            {
                $command.Dispose();
            }
        }
    }
    catch [System.Data.SqlClient.SqlException]
    {
        if ([System.String]::IsNullOrEmpty($ErrorMessages) -eq $false)
        {
            $ErrorMessages += "`r`n";
        }
        $ErrorMessages += $_.Exception.Message;
    }
    finally
    {
        if ($null -ne $connection)
        {
            $connection.Close();
        }
    }

    if ([System.String]::IsNullOrEmpty($ErrorMessages) -eq $false)
    {
        $ErrorMessages = "Error restoring file $BackupFileName to database $DatabaseName to instance $InstanceName : " + $ErrorMessages;
        throw $ErrorMessages;
    }

    return $restoreScriptSql;
}

####################################################################################################
## Backup-Database
####################################################################################################
function Backup-Database
{
    param
    (
        [string] $InstanceName,
        [string[]] $DatabaseList,
        [string] $BackupType = "Full",
        [string] $BackupPath,
        [string] $FileExtension = "bak",
        [bool] $AppendInstanceToBackupPath = $true,
        [bool] $PrependInstanceToBackupName = $true,
        [bool] $UseCompression = $true,
        [bool] $CopyOnly = $false,
        [bool] $ContinueOnFailure = $true,
        [int] $BackupTimeoutSeconds = 600
    )

    $ErrorMessages = "";

    $instanceBackupPath = $BackupPath;
    if ($AppendInstanceToBackupPath -eq $true)
    {
        [string[]]$instanceParts = $InstanceName.Split("\");
        if ($instanceParts.Count -lt 2)
        {
            $instanceParts += "MSSQLSERVER";
        }

        $instanceBackupPath = [System.IO.Path]::Combine($BackupPath, $instanceParts[0]);
        $instanceBackupPath = [System.IO.Path]::Combine($instanceBackupPath, $instanceParts[1]);
    }

    if ($UseCompression -eq $true)
    {
        $compression = ", compression";
    }
    else
    {
        $compression = "";
    }

    if ($CopyOnly -eq $true)
    {
        $copyOnlyClause = ", copy_only";
    }
    else
    {
        $copyOnlyClause = "";
    }

    $sb = New-Object System.Data.SqlClient.SqlConnectionStringBuilder;
    $sb["Data Source"] = $InstanceName;
    $sb["Integrated Security"] = $true;

    try
    {
        $connection = New-Object System.Data.SqlClient.SqlConnection($sb.ToString());
        $connection.Open();

        if ($BackupType -eq "Full")
        {
            $commandType = "database";
            $differential = "";
            $fileSuffix = "Full"
        }
        elseif ($BackupType -eq "Diff" -or $BackupType -eq "Differential")
        {
            $commandType = "database";
            $differential = ", differential";
            $fileSuffix = "Diff";
        }
        elseif ($BackupType -eq "Log")
        {
            $commandType = "log";
            $differential = "";
            $fileSuffix = "Log";
        }
        else
        {
            throw "Invalid BackupType specified: $BackupType";
        }

        foreach ($database in $DatabaseList)
        {
            $timestamp = (Get-Date).ToString("yyyyMMddHHmmss");
            $backupNamePrefix = "";
            if ($PrependInstanceToBackupName -eq $true)
            {
                $backupNamePrefix = $InstanceName.Replace("\", "_") + "_";
            }
            $backupName = $backupNamePrefix + $database + $fileSuffix + $timestamp + "." + $FileExtension;
            $databaseBackupPath = [System.IO.Path]::Combine($instanceBackupPath, $database);
            $backupFileName = [System.IO.Path]::Combine($databaseBackupPath, $backupName);
            $backupFileName = $backupFileName.Replace("'", "''");

            $sql = "backup $commandType [$database] to disk = '$backupFileName' with init, skip, norewind, nounload, stats = 10, buffercount = 30 $differential $compression $copyOnlyClause;";

            [void][System.IO.Directory]::CreateDirectory($databaseBackupPath);

            try
            {
                $command = New-Object System.Data.SqlClient.SqlCommand($sql, $connection);
                $command.CommandTimeout = $BackupTimeoutSeconds;
                [void]$command.ExecuteNonQuery();
            }
            catch [Exception]
            {
                if ($ContinueOnFailure -eq $true)
                {
                    if ([System.String]::IsNullOrEmpty($ErrorMessages) -eq $false)
                    {
                        $ErrorMessages += "`r`n";
                    }
                    $ErrorMessages += $_.Exception.Message;
                }
                else
                {
                    throw;
                }
            }
            finally
            {
                if ($null -ne $command)
                {
                    $command.Dispose();
                }
            }
        }
    }
    catch [System.Data.SqlClient.SqlException]
    {
        if ([System.String]::IsNullOrEmpty($ErrorMessages) -eq $false)
        {
            $ErrorMessages += "`r`n";
        }
        $ErrorMessages += $_.Exception.Message;
    }
    finally
    {
        if ($null -ne $connection)
        {
            $connection.Close();
        }
    }

    if ($ContinueOnFailure -and [System.String]::IsNullOrEmpty($ErrorMessages) -eq $false)
    {
        throw $ErrorMessages;
    }
}

####################################################################################################
## Invoke-VerifyMostRecentBackup
####################################################################################################
function Invoke-VerifyMostRecentBackup
{
    param
    (
        [Parameter(Mandatory = $true)]
        [string] $SourcePath,

        [Parameter(Mandatory = $true)]
        [string[]] $DatabaseList,

        [Parameter(Mandatory = $true)]
        [string] $BackupType,

        [Parameter(Mandatory = $false)]
        [string] $FileExtension = "bak",

        [Parameter(Mandatory = $true)]
        [string] $RestoreServerName,

        [Parameter(Mandatory = $false)]
        [int] $VerifyTimeoutSeconds = 3600,

        [Parameter(Mandatory = $false)]
        [bool] $ContinueOnFailure = $true
    )

    $ErrorMessages = "";

    $sb = New-Object System.Data.SqlClient.SqlConnectionStringBuilder;
    $sb["Data Source"] = $RestoreServerName;
    $sb["Integrated Security"] = $true;

    $sql = "restore verifyonly from disk = @fileName;";

    try
    {
        $connection = New-Object System.Data.SqlClient.SqlConnection($sb.ToString());
        $connection.Open();

        foreach($db in $DatabaseList)
        {
            $pattern = "*$db*$BackupType*.$FileExtension";
            $searchPath = [System.IO.Path]::Combine($SourcePath, $db);

            $file = Get-ChildItem -Path $searchPath -Filter $pattern | Where-Object { $_.LastWriteTime -ge  $purgeFromTime} | Sort-Object -Property LastWriteTime -Descending | Select-Object -First 1;

            if ($null -ne $file)
            {
                try
                {
                    $command = New-Object System.Data.SqlClient.SqlCommand($sql, $connection);
                    [void]$command.Parameters.Add("@fileName", $file.FullName);
                    $command.CommandTimeout = $VerifyTimeoutSeconds;
                    $command.ExecuteNonQuery();
                }
                catch [System.Data.SqlClient.SqlException]
                {
                    $ex = $_.Exception;
                    if ($ContinueOnFailure)
                    {
                        if ([System.String]::IsNullOrEmpty($ErrorMessages) -eq $false)
                        {
                            $ErrorMessages += "`r`n";
                        }
                        $ErrorMessages += $ex.Message;
                    }
                    else
                    {
                        throw $ex;
                    }
                }
                finally
                {
                    $command.Dispose();
                }
            }
        }
    }
    catch [System.Exception]
    {
        $ex = $_.Exception;
        if ($ContinueOnFailure)
        {
            if ([System.String]::IsNullOrEmpty($ErrorMessages) -eq $false)
            {
                $ErrorMessages += "`r`n";
            }
            $ErrorMessages += $ex.Message;
        }
        else
        {
            throw $ex;
        }
    }
    finally
    {
        if ($null -ne $connection)
        {
            if ($connection.State -eq [System.Data.ConnectionState]::Open)
            {
                $connection.Close();
            }
            else
            {
                $connection.Dispose();
            }
        }
    }

    if ($ContinueOnFailure -and [System.String]::IsNullOrEmpty($ErrorMessages) -eq $false)
    {
        throw $ErrorMessages;
    }
}

####################################################################################################
## Invoke-CheckDB
####################################################################################################
function Invoke-CheckDB
{
    param
    (
        [string] $InstanceName,
        [string[]] $DatabaseList,
        [bool] $ContinueOnFailure = $true,
        [int] $CheckDBTimeoutSeconds = 600
    )

    $sb = New-Object System.Data.SqlClient.SqlConnectionStringBuilder;
    $sb["Data Source"] = $InstanceName;
    $sb["Integrated Security"] = $true;

    try
    {
        $connection = New-Object System.Data.SqlClient.SqlConnection($sb.ToString());
        $connection.Open();

        foreach ($database in $DatabaseList)
        {
            $sql = "dbcc checkdb ([$database]) with no_infomsgs, all_errormsgs;";

            try
            {
                $command = New-Object System.Data.SqlClient.SqlCommand($sql, $connection);
                $command.CommandTimeout = $BackupTimeoutSeconds;
                $command.ExecuteNonQuery();
            }
            catch [Exception]
            {
                if ($ContinueOnFailure -eq $true)
                {
                    if ([System.String]::IsNullOrEmpty($ErrorMessages) -eq $false)
                    {
                        $ErrorMessages += "`r`n";
                    }
                    $ErrorMessages += $_.Exception.Message;
                }
                else
                {
                    throw;
                }
            }
            finally
            {
                if ($null -ne $command)
                {
                    $command.Dispose();
                }
            }
        }
    }
    catch [System.Data.SqlClient.SqlException]
    {
        if ([System.String]::IsNullOrEmpty($ErrorMessages) -eq $false)
        {
            $ErrorMessages += "`r`n";
        }
        $ErrorMessages += $_.Exception.Message;
    }
    finally
    {
        if ($null -ne $connection)
        {
            $connection.Close();
        }
    }

    if ($ContinueOnFailure -and [System.String]::IsNullOrEmpty($ErrorMessages) -eq $false)
    {
        throw $ErrorMessages;
    }
}

####################################################################################################
## Get-RestoreChain
####################################################################################################
function Get-RestoreChain
{
    param
    (
        [Parameter(Mandatory = $true)]
        [string] $SourceFullBackupPath,

        [Parameter(Mandatory = $true)]
        [string] $FullPattern,

        [string] $SourceLogBackupPath = $null,
        [string] $LogPattern = $null,

        [Parameter(Mandatory = $true)]
        [string] $RestoreServerName,

        [Parameter(Mandatory = $true)]
        [DateTime] $RestorePoint,

        [string] $BackupApplication = "Native"
    )

    if ([System.String]::IsNullOrEmpty($SourceLogBackupPath))
    {
        $SourceLogBackupPath = $SourceFullBackupPath;
    }

    ## Create the Sql.RestoreInfo type
    Add-CustomTypesInternal;

    ## Get the backup files in the directories
    $backupFiles = New-Object System.Collections.Generic.List[System.IO.FileInfo];

    $fullFiles = (New-Object System.IO.DirectoryInfo($SourceFullBackupPath)).GetFiles($FullPattern);
    $backupFiles.AddRange($fullFiles);

    if ([System.String]::IsNullOrEmpty($LogPattern) -eq $false)
    {
        $logFiles = (New-Object System.IO.DirectoryInfo($SourceLogBackupPath)).GetFiles($LogPattern);
        $backupFiles.AddRange($logFiles);
    }

    ## Read cached restore info data
    $restoreInfoXml = [System.IO.Path]::Combine($SourceFullBackupPath, "HeaderOnlyInfo.xml");
    $cachedRestoreInfoList = New-Object Sql.RestoreInfoList;
    if ([System.IO.File]::Exists($restoreInfoXml))
    {
        $reader = New-Object System.IO.StreamReader($restoreInfoXml);
        try
        {
            $serializer = New-Object System.Xml.Serialization.XmlSerializer($cachedRestoreInfoList.GetType());
            $cachedRestoreInfoList = $serializer.Deserialize($reader);
        }
        finally
        {
            if ($null -ne $reader)
            {
                $reader.Dispose();
                $reader = $null;
            }
        }
    }

    ## Read the restore information for each backup file
    $restoreInfoList = New-Object Sql.RestoreInfoList;
    foreach ($backupFile in $backupFiles)
    {
        $backupFileName = $backupFile.FullName;
        $cachedRestoreInfo = $cachedRestoreInfoList.GetRestoreInfo($backupFileName);
        if ($null -eq $cachedRestoreInfo)
        {
            $restoreInfo = Get-RestoreInfo -BackupFileName $backupFileName -BackupApplication $BackupApplication -RestoreServerName $RestoreServerName;
            $restoreInfoList.Add($restoreInfo);
        }
        else
        {
            $restoreInfoList.Add($cachedRestoreInfo);
        }
    }

    ## Cache the restore information
    $writer = New-Object System.IO.StreamWriter($restoreInfoXml);
    try
    {
        $serializer = New-Object System.Xml.Serialization.XmlSerializer($restoreInfoList.GetType());
        $serializer.Serialize($writer, $restoreInfoList);
    }
    finally
    {
        if ($null -ne $writer)
        {
            $writer.Dispose();
            $writer = $null;
        }
    }

    ## Get latest backup before the requested start time.
    $backupsToRestore = New-Object Sql.RestoreInfoList;
    $full = $restoreInfoList |
        Where-Object { $_.BackupType -eq "Full" -and $_.BackupStartDate -le $RestorePoint } |
        Sort-Object -Property BackupStartDate -Descending |
        Select-Object -First 1;
    if ($null -ne $full)
    {
        $backupsToRestore.Add($full);
    }

    ## Get log backups after the full up to the requested start time
    $logsList = New-Object Sql.RestoreInfoList;
    $restoreInfoList |
            Where-Object { $_.BackupType -eq "Log" -and $_.BackupStartDate -le $RestorePoint -and $_.LastLsn -ge $full.LastLsn } |
            Sort-Object -Property FirstLSN, BackupStartDate |
            ForEach-Object { $logsList.Add($_) };

    if ($null -ne $logsList -and $logsList.Count -gt 0)
    {
        $backupsToRestore.AddRange($logsList);
    }

    ## Still need one more log to catch up to the restore point
    if ($backupsToRestore.Count -gt 0)
    {
        $lastLog = $restoreInfoList |
            Where-Object { $_.BackupType -eq "Log" -and $_.BackupStartDate -gt $backupsToRestore[$backupsToRestore.Count - 1].BackupStartDate } |
            Sort-Object -Property BackupStartDate |
            Select-Object -First 1;
        if ($null -ne $lastLog)
        {
            $backupsToRestore.Add($lastLog);
        }
    }

    return $backupsToRestore;
}

####################################################################################################
## Restore-DatabaseFullOnly
####################################################################################################
function Restore-DatabaseFullOnly
{
    param
    (
        [Parameter(Mandatory = $true)]
        [string] $DatabaseName,

        [Parameter(Mandatory = $true)]
        [string] $SourceFullBackupPath,

        [Parameter(Mandatory = $true)]
        [string] $FullPattern,

        [Parameter(Mandatory = $true)]
        [string] $RestoreServerName,

        [string] $MdfPathName,
        [string] $LdfPathName,

        [Nullable[DateTime]] $RestorePoint = $null,

        [string] $BackupApplication = "Native",
        [bool] $SetToSimple = $false,
        [bool] $OutputScriptOnly = $false
    )

    if ($null -eq $RestorePoint)
    {
        $RestorePoint = Get-Date;
    }

    $restoreChain = Get-RestoreChain `
        -SourceFullBackupPath $SourceFullBackupPath `
        -FullPattern $FullPattern `
        -RestoreServerName $RestoreServerName `
        -RestorePoint $RestorePoint `
        -BackupApplication $BackupApplication;

    if ($null -eq $restoreChain -or $restoreChain.Length -eq 0)
    {
        throw "Unable to determine restore chain to complete restore process.";
    }

    if ($restoreChain.Length -gt 1)
    {
        throw "More than one backup file identified to restore.";
    }

    ## Restore the full
    $restoreSqlScript += Restore-DatabaseInternal `
        -InstanceName $restoreServerName `
        -DatabaseName $databaseName `
        -BackupFileName $restoreChain.BackupFileName `
        -MdfPathName $MdfPathName `
        -LdfPathName $LdfPathName `
        -BackupType "Full" `
        -OutputScriptOnly $OutputScriptOnly;

    ## Recover the database
    $restoreSqlScript += Restore-DatabaseInternal `
        -InstanceName $restoreServerName `
        -DatabaseName $databaseName `
        -BackupType "Recovery" `
        -SetToSimple $SetToSimple `
        -OutputScriptOnly $OutputScriptOnly;

    return $restoreSqlScript;
}

####################################################################################################
## Restore-DatabaseFullAndLogs
####################################################################################################
function Restore-DatabaseFullAndLogs
{
    param
    (
        [Parameter(Mandatory = $true)]
        [string] $DatabaseName,

        [Parameter(Mandatory = $true)]
        [string] $SourceFullBackupPath,

        [Parameter(Mandatory = $true)]
        [string] $FullPattern,

        [Parameter(Mandatory = $true)]
        [string] $SourceLogBackupPath,

        [Parameter(Mandatory = $true)]
        [string] $LogPattern,

        [Parameter(Mandatory = $true)]
        [string] $RestoreServerName,

        [string] $MdfPathName,

        [string] $LdfPathName = $null,

        [Parameter(Mandatory = $true)]
        [DateTime] $RestorePoint = $null,

        [string] $BackupApplication = "Native",
        [bool] $SetToSimple = $false,
        [bool] $OutputScriptOnly = $false
    )

    $restoreScriptSql = "";

    $restoreChain = Get-RestoreChain `
        -SourceFullBackupPath $SourceFullBackupPath `
        -FullPattern $FullPattern `
        -SourceLogBackupPath $SourceLogBackupPath `
        -LogPattern $LogPattern `
        -RestoreServerName $RestoreServerName `
        -RestorePoint $RestorePoint `
        -BackupApplication $BackupApplication;

    if ($null -eq $restoreChain -or $restoreChain.Length -eq 0)
    {
        throw "Unable to determine restore chain to complete restore process.";
    }

    if ([System.String]::IsNullOrEmpty($LdfPathName))
    {
        $LdfPathName = $MdfPathName;
    }

    ## Restore the full and logs
    for($i = 0; $i -lt $restoreChain.Count; $i++)
    {
        $backup = $restoreChain[$i];

        if ($i -eq 0)
        {
            $restoreScriptSql += Restore-DatabaseInternal `
                -InstanceName $restoreServerName `
                -DatabaseName $databaseName `
                -BackupFileName $backup.BackupFileName `
                -MdfPathName $MdfPathName `
                -LdfPathName $LdfPathName `
                -BackupType "Full" `
                -OutputScriptOnly $OutputScriptOnly;
        }
        elseif ($i -lt $restoreChain.Count - 1)
        {
            $restoreScriptSql += Restore-DatabaseInternal `
                -InstanceName $restoreServerName `
                -DatabaseName $databaseName `
                -BackupFileName $backup.BackupFileName `
                -BackupType "Log" `
                -OutputScriptOnly $OutputScriptOnly;
        }
        else
        {
            $restoreScriptSql += Restore-DatabaseInternal `
                -InstanceName $restoreServerName `
                -DatabaseName $databaseName `
                -BackupFileName $backup.BackupFileName `
                -BackupType "Log" `
                -RestorePoint $restorePoint `
                -OutputScriptOnly $OutputScriptOnly;
        }
    }

    ## Recover the database
    $restoreScriptSql += Restore-DatabaseInternal `
        -InstanceName $restoreServerName `
        -DatabaseName $databaseName `
        -BackupType "Recovery" `
        -SetToSimple $SetToSimple `
        -OutputScriptOnly $OutputScriptOnly;

    return $restoreScriptSql;
}

####################################################################################################
## Get-LastRebootTime
####################################################################################################
function Get-LastRebootTime
{
<#
.SYNOPSIS

Determine the last time that a machine was rebooted.

.DESCRIPTION

Calculate and return the last time that the input machine was rebooted.

.PARAMETER ComputerName

The name of the computer.

.PARAMETER ThrowExceptionOnError

If $true, an exception will be thrown if an error occurs. If $false (default), the reboot time will be returned as blank.

.INPUTS

An array of strings may be piped to this parameters. Each string represents the ComputerName parameter values.

.OUTPUTS

An object containing the computer name and the last reboot time of the computer.

.EXAMPLE

Get-LastRebootTime -ComputerName "localhost" -ThrowExceptionOnError $true;

.EXAMPLE

"server1", "server2" | Get-LastRebootTime;

#>
    [CmdletBinding()]

    param
    (
        [Parameter(Mandatory = $true, ValueFromPipeline = $true)]
        [string] $ComputerName,

        [bool] $ThrowExceptionOnError = $false
    )

    process
    {
        try
        {
            $lastBoot = Get-WmiObject Win32_OperatingSystem -ComputerName $ComputerName | Select-Object @{label="LastBootUpTime";expression={$_.ConvertToDateTime($_.LastBootUpTime)}};
        }
        catch
        {
            if ($ThrowExceptionOnError -eq $true)
            {
                throw;
            }

            $lastBoot = New-Object System.DateTime;
        }

        $bootInfo = New-Object -TypeName psobject;
        $bootInfo | Add-Member -MemberType NoteProperty -Name ComputerName -Value $ComputerName;
        $bootInfo | Add-Member -MemberType NoteProperty -Name LastBootTime -Value $lastBoot.LastBootupTime;

        return $bootInfo;
    }
}

####################################################################################################
